// StrapDisplay 2.13" E-Paper & 3x ThinPot Prototype
//-------------------------------------------------------------------------
// Board: E-Paper ESP8266 Driver Board [Arduino Board: NodeMCU 1.0 (ESP-12E Module)]
// Display: WaveShare 2.13d, Flexible B/W E-Paper
// Analog-Digital-Converter: ADS 1115 via I2C (Pin: D6,D7)
// Sensors: 3x ThinPot 50mm (Pin: A0-A2, ADS115)

// Includes
//------------------------------------
#include <Wire.h>           // Library for I2C communication
#include <ESP8266WiFi.h>
#include <ESP8266WiFiMulti.h>
#include <ArduinoJson.h>      // Library for JSON format
#include <WebSocketsClient.h> // Library for WebSockets
#include <SocketIOclient.h>   // SocketIO implementation for Arduino
#include "Adafruit_ADS1015.h" // Library for I2C ADS1115 sensor
#include <GxEPD2_BW.h>

// Defines
//------------------------------------
#define WEBSOCKETS_NETWORK_TYPE NETWORK_ESP8266_ASYNC   // 
//#define STRAP_TYPE "LOWER_STRAP"
#define STRAP_TYPE "UPPER_STRAP"

// SoftPot Calibration
//------------------------------------

// Sensor pins
#define rightSoftPotPin 0
#define middleSoftPotPin 1
#define leftSoftPotPin 2

// SPI pin definition
#define CS_PIN           15
#define RST_PIN          5
#define DC_PIN           4
#define BUSY_PIN         16

// Sensor ranges

// Lower Strap Calibration
int leftSoftPotMin = 8900;
int leftSoftPotMax = 16000;
int middleSoftPotMin = 550;
int middleSoftPotMax = 15300;
int rightSoftPotMin = 500;
int rightSoftPotMax = 15300;

/*
// Upper Strap Calibration
int leftSoftPotMin = 2500;
int leftSoftPotMax = 14600;
int middleSoftPotMin = 2000;
int middleSoftPotMax = 15000;
int rightSoftPotMin = 2000;
int rightSoftPotMax = 15000;
*/

// Resolution & filter threshold
int sensingRes = 10;          // bit; e.g., 10 = 2^10 [=0-1023]
int moveThreshold = 30;

// Serial Debug Configuration
//------------------------------------
bool show_JSON_Serial = true;
bool show_SensorDataEvents_Serial = false;
bool show_SensorData_Serial = false;
bool show_RawSensorData_Serial = false;

// General Configuration
//------------------------------------
const char* wifi_ssid = "IMLDLab";
const char* wifi_password = "deadl1ne13";
const char* socketIO_ServerIP = "141.76.67.139";
const int socketIO_ServerPort = 31415;
bool send_InputEventsViaSocketIO = true;

// Global Vars
//------------------------------------
Adafruit_ADS1115 ads(0x48);   // Analog-To-Digital converter (ADDR = GND --> 0x48 I2C)
SocketIOclient socketIO;      // SocketIO object
unsigned char image[2756];
GxEPD2_BW<GxEPD2_213_flex, GxEPD2_213_flex::HEIGHT> display(GxEPD2_213_flex(CS_PIN, DC_PIN, RST_PIN, BUSY_PIN));

// Sensor Data Vars
//------------------------------------
int16_t rightSoftPotRawValue, middleSoftPotRawValue, leftSoftPotRawValue;       // Captured RAW values from the SoftPots sensors
int16_t rightSoftPotValue,  middleSoftPotValue,  leftSoftPotValue;              // Filtered & mapped values from the SoftPots sensors
int16_t prevRightSoftPotValue,  prevMiddleSoftPotValue,  prevLeftSoftPotValue;  // Previous values from the SoftPots sensors [filtered and mapped]
int16_t sensorDatabase10x[10][3];                                               // saves the last ten sensor values of all three soft potentiometer sensors
int sensorDatabaseCounter;                                                      // countinously count from 0 to 9 to access the sensor data array in time
int16_t validPreMoveVal;                                                        // valid (=outside filter threshold) value of the previous touch movement event

// State Machine Vars
//------------------------------------
bool EPD_Update_Busy = false;
bool touchDownBegin[3];
bool touchDown[3];
int touchDownBeginCounter[3];

// WiFi
//------------------------------------
IPAddress myIP;

// Misc
//------------------------------------
uint64_t now;
unsigned long messageTimestamp = 0;

// Setup
//------------------------------------
void setup(void)
{
  // For Debug
  Serial.begin(115200);

  // Configure PGA of ADS 115
  // 2/3x gain +/- 6.144V  1 bit = 3mV      0.1875mV
  ads.setGain(GAIN_TWOTHIRDS);

  
  WiFi.mode(WIFI_STA);
  WiFi.begin(wifi_ssid, wifi_password);
  delay(1000);
  Serial.println();
  Serial.print("MAC: ");
  Serial.println(WiFi.macAddress());

  // Connect to WiFi network
  Serial.println("");
  Serial.println("");
  Serial.print("Connecting to ");
  Serial.println(wifi_ssid);

  // SPI initialization
  pinMode(CS_PIN  , OUTPUT);
  pinMode(RST_PIN , OUTPUT);
  pinMode(DC_PIN  , OUTPUT);
  pinMode(BUSY_PIN,  INPUT);
  SPI.begin();

  // Wait for connection
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }

  // Display IP address...
  delay(300);
  Serial.println("\nConnected.");
  Serial.print("IP address: ");
  Serial.println(WiFi.localIP());
  delay(300);

  socketIO.begin(socketIO_ServerIP, socketIO_ServerPort);

  display.init(115200);

  socketIO.onEvent(socketIOEvent);

  // Start ADS 1115
  ads.begin();
}

// Loop
//------------------------------------
void loop(void)
{
  // Handle SocketIO server...
  socketIO.loop();

  // Handle touch input...
  handleTouchInput();
}

// Helper Functions
//------------------------------------
void socketIOEvent(socketIOmessageType_t type, uint8_t * payload, size_t length) {

  switch (type) {

    case sIOtype_CONNECT: // = 48
      sendHandshake(show_JSON_Serial);
      break;

    case sIOtype_DISCONNECT: // = 49
      break;

    case sIOtype_EVENT: // = 50

      // Output event message...
      if (show_JSON_Serial) {
        Serial.printf("%s\n", payload);
      }

      // Create dynamic JSON document...
      const size_t capacity = 18000; // the imageData messages are 14800. 3200 buffer.
      DynamicJsonDocument doc(capacity);

      // Try to deserialize the recevied JSON message...
      //DeserializationError error = deserializeJson(doc, payload, length); // ToDo: replace

      DeserializationError error = deserializeJson(doc, payload, length + 100);
      if (error) {
        Serial.print("deserializeJson() failed: ");
        Serial.println(error.c_str());
        return;
      }
      
      // Debug memory usage...
      // Serial.print("MemoryUsage: ");
      // Serial.print(doc.memoryUsage());
      // Serial.print("/");
      // Serial.println(doc.capacity());

      // Get event name from message (e.g., "handshake" or "msg")...
      String eventName = doc[0];

      // Switch event name....
      //----------------------
      // >> "handshake" event
      if (eventName == "handshake") {}

      // >> "msg" event
      else if (eventName == "msg") {

        // Get "part" value and cast it to int...
        String strPart = doc[1]["part"];
        int intPart = strPart.toInt();

        // Get "payload" value
        JsonArray dataArray = doc[1]["payload"];

        // "payload" explanation
        // -------------------------------------------------------------
        // = 918 Bytes (for intPart==0), 919 Bytes (intPart==1 || intPart==2)
        // = (918+919+919) = 2756 Bytes in total
        // = 2756 * 8 = 22048 bits
        // = 22048 bits = 212px x 104px Display Resolution
        // Example Transmission:
        // Bytes:                  Image (one row)
        // 00110000 00110000  =>   |  x x    |  x x    |
        // -------------------------------------------------------------


        // Fill EPD_image_buffer with pixels ("payload") of the
        // current package ("part"). Note that only the last
        // package ("part":"2") triggers the E-Ink update.
        // Therefore, the whole EPD_image_buffer is updated
        // -------------------------------------------------------------
        // ToDo: double check correct index
        // if intPart==0 918
        // if intPart==1 919
        // if intPart==2 919
        if (intPart == 0) {
          for (int i = 0; i < 918; i++) {
            const unsigned char data = dataArray[i];
            image[i] = data;
          }
        }
        else if (intPart == 1) {
          for (int i = 918; i < 1837; i++) {
            const unsigned char data = dataArray[i - 918];
            image[i] = data;
          }
        }
        else if (intPart == 2) {
          for (int i = 1837; i < 2756; i++) {
            const unsigned char data = dataArray[i - (2 * 918) - 1];
            image[i] = data;
          }
        }

        // If it is the last package upload the
        // previously filled EPD_image_buffer
        if (intPart == 2) {
          // if(intPart==2 && !loading){ // ToDo: replace later
          EPD_Update_Busy = true;

          Wire.endTransmission();
          SPI.begin();
          delay(10);

          if (doc[1]["refresh"]=="full"){
            fullRefresh(image);
          }
          else if(doc[1]["refresh"]=="part"){
            partialRefresh(image); 
          }
          
          delay(10);
          SPI.end();
          delay(10);
        
          // Reinit I2C ADC-board
          ads.reinit(0x48);
          EPD_Update_Busy = false;
        }

      } // end message
      break;


  } // End: switch(type=
} // End: void socketIOEvent(socketIOmessageType_t type, uint8_t * payload, size_t length)

/*  handleTouchInput()
   ------------------
   ...
*/
void handleTouchInput() {

  // If the ESP8266 currently flashes the e-paper display
  // deactive any SoftPot sensing activities...
  if (EPD_Update_Busy == false) {

    // Read sensor values...
    rightSoftPotRawValue = ads.readADC_SingleEnded(rightSoftPotPin);
    middleSoftPotRawValue = ads.readADC_SingleEnded(middleSoftPotPin);
    leftSoftPotRawValue = ads.readADC_SingleEnded(leftSoftPotPin);

    // Filter & normalize sensor values

    // > Sensor #01
    cutExtrema(&leftSoftPotRawValue, leftSoftPotMin, leftSoftPotMax);
    leftSoftPotValue = map(leftSoftPotRawValue, leftSoftPotMin, leftSoftPotMax, 0, pow(2, sensingRes) - 1);

    // > Sensor #02
    cutExtrema(&middleSoftPotRawValue, middleSoftPotMin, middleSoftPotMax);
    middleSoftPotValue = map(middleSoftPotRawValue, middleSoftPotMin, middleSoftPotMax, 0, pow(2, sensingRes) - 1);

    // > Sensor #03
    cutExtrema(&rightSoftPotRawValue, rightSoftPotMin, rightSoftPotMax);
    rightSoftPotValue = map(rightSoftPotRawValue, rightSoftPotMin, rightSoftPotMax, 0, pow(2, sensingRes) - 1);

    // Since the analog-to-digital converter has some multiplexing problems,
    // we focus on the highest values that indicates the current touch.
    // Therefore, we cut the lower lower (noisy) values from the inactive slides.
    zeroMinimumValues(&rightSoftPotValue, &middleSoftPotValue, &leftSoftPotValue);

    // Show RAW sensor data...
    if (show_RawSensorData_Serial) {
      Serial.print(leftSoftPotRawValue);
      Serial.print("/");
      Serial.print(middleSoftPotRawValue);
      Serial.print("/");
      Serial.println(rightSoftPotRawValue);
    }

    // Show filtered and mapped sensor data...
    if (show_SensorData_Serial) {
      Serial.print(leftSoftPotValue);
      Serial.print("/");
      Serial.print(middleSoftPotValue);
      Serial.print("/");
      Serial.println(rightSoftPotValue);
    }

    // Check touch input
    checkTouch(1, prevLeftSoftPotValue, leftSoftPotValue, moveThreshold, send_InputEventsViaSocketIO, show_SensorDataEvents_Serial);
    checkTouch(2, prevMiddleSoftPotValue, middleSoftPotValue, moveThreshold, send_InputEventsViaSocketIO, show_SensorDataEvents_Serial);
    checkTouch(3, prevRightSoftPotValue, rightSoftPotValue, moveThreshold, send_InputEventsViaSocketIO, show_SensorDataEvents_Serial);

    // Store sensor values for direct comparison in the next loop
    prevRightSoftPotValue = rightSoftPotValue;
    prevMiddleSoftPotValue = middleSoftPotValue;
    prevLeftSoftPotValue = leftSoftPotValue;

    // Store sensor values in database for advanced event recognition....
    storeInSensorDatabase10x(leftSoftPotValue, middleSoftPotValue, rightSoftPotValue);
  }
}

/*  storeInSensorDatabase10x()
   ------------------
   ...
*/
void storeInSensorDatabase10x(int16_t leftSoftPotValue, int16_t middleSoftPotValue, int16_t rightSoftPotValue) {

  if (sensorDatabaseCounter > 9) {
    sensorDatabaseCounter = 0;
  }

  sensorDatabase10x[sensorDatabaseCounter][0] = leftSoftPotValue;
  sensorDatabase10x[sensorDatabaseCounter][1] = middleSoftPotValue;
  sensorDatabase10x[sensorDatabaseCounter][2] = rightSoftPotValue;

  sensorDatabaseCounter++;
}

/*  getMaxHistoryValue()
   ------------------
   ...
*/
int16_t getMaxHistoryValue(int16_t touch_x, int16_t last) {
  int currentMax = 0;
  for (int i = last; i >= 0; i--) {
    int index = ((10 + (sensorDatabaseCounter - i)) % 10);
    int yVal = sensorDatabase10x[index][touch_x - 1];
    if ( yVal > currentMax) {
      currentMax = yVal;
    }
  }
  return currentMax;
}

/*  getMinHistoryValue()
    ------------------
    ...
*/
int16_t getMinHistoryValue(int16_t touch_x, int16_t last) {
  int currentMin = 1023;
  for (int i = last; i >= 0; i--) {
    int index = ((10 + (sensorDatabaseCounter - i)) % 10);
    int yVal = sensorDatabase10x[index][touch_x - 1];
    if ( yVal < currentMin) {
      currentMin = yVal;
    }
  }
  return currentMin;
}

/*  sendTouchEvent()
    ------------------
    @param touch_type defines the type of the touch event such as touchstart, touchmove or touchend
    @param touch_x the id of the SoftPot sensor representing the x value.
    @param touch_y the captured sensor value representing the y value.
    @param serialDebug indicates whether the function should print the event data on the serial console
*/
void sendTouchEvent(String touch_type, int16_t touch_x, int16_t touch_y, bool serialDebug) {

  DynamicJsonDocument jsonBuffer(1024);
  JsonArray array = jsonBuffer.to<JsonArray>();

  array.add("msg");
  JsonObject json_event = array.createNestedObject();
  json_event["target"]  = "MAIN";
  json_event["type"]  = "softpottouch";
  json_event["sender"]  = STRAP_TYPE;
  JsonObject json_payload = json_event.createNestedObject("payload");
  json_payload["type"]    = touch_type;
  json_payload["timestamp"]    = millis();
  JsonObject json_position = json_payload.createNestedObject("pos");
  json_position["x"] = touch_x;
  json_position["y"] = touch_y;

  // JSON to String (serializion)
  String output;
  serializeJson(jsonBuffer, output);

  // Serial debug
  if (serialDebug) {
    Serial.println(output);
  }

  // Send event
  socketIO.sendEVENT(output);

  return;
}

void sendHandshake(bool serialDebug) {

  DynamicJsonDocument jsonBuffer(1024);
  JsonArray array = jsonBuffer.to<JsonArray>();

  array.add("handshake");
  array.add(STRAP_TYPE);

  // JSON to String (serializion)
  String output;
  serializeJson(jsonBuffer, output);

  // Serial debug
  if (serialDebug) {
    Serial.println(output);
  }

  // Send event
  socketIO.sendEVENT(output);

  return;
}

/*  checkTouch()
    ------------------
    This function captures a single SoftPot sensor and manages the respective event handling via websockets.
*/
void checkTouch (int16_t touch_x, int16_t prevSoftPotValue, int16_t softPotValue, int16_t threshold, bool sendEvent, bool serialDebug) {

  if (touchDownBegin[touch_x - 1] &&  (touchDown[touch_x - 1] == false)) {

    if (touchDownBeginCounter[touch_x - 1] > 3) {

      // Filter touchstart events during a touchevent already exists...
      if (getMaxHistoryValue(touch_x, 3) == 0) {
        touchDownBegin[touch_x - 1] = false;
        touchDownBeginCounter[touch_x - 1] = 0;
        return;
      }

      // Serial debug
      if (serialDebug == true) {
        Serial.print("Start at position: ");
        Serial.print(touch_x);
        Serial.print(" / ");
        Serial.println(getMaxHistoryValue(touch_x, 3));
      }

      // Handle state machines
      touchDown[touch_x - 1] = true;
      validPreMoveVal = getMaxHistoryValue(touch_x, 3);
      touchDownBegin[touch_x - 1] = false;
      touchDownBeginCounter[touch_x - 1] = 0;

      // Send Event via SocketIO
      if (sendEvent) {
        sendTouchEvent("touchstart", touch_x, getMaxHistoryValue(touch_x, 3), show_JSON_Serial);
      }
    }

    // Increase touch down begin counter to store values
    touchDownBeginCounter[touch_x - 1]++;
    return;
  }

  // Touch Start
  // ------------------
  // Identification criteria: pre-value == 0 + currentValue > 0 + no other TouchStart
  if ((prevSoftPotValue == 0) && (softPotValue > 0) && (touchDownBegin[touch_x - 1] == false)) {

    // Set touchDownBegin high and the touchDownBeginCounter to zero
    // to start the touch event handling loop
    touchDownBegin[touch_x - 1] = true;
    touchDownBeginCounter[touch_x - 1] = 0;

    // We expect no other touch events in this run
    return;
  }

  // Touch End
  // ------------------
  else if (((prevSoftPotValue != 0)) && (softPotValue == 0)) {

    // Serial debug
    if (serialDebug == true) {
      Serial.print("End at position: ");
      Serial.print(touch_x);
      Serial.print(" / ");
      Serial.println(getMaxHistoryValue(touch_x, 3));
    }

    // Handle state machines
    touchDownBegin[touch_x - 1] = false;
    touchDownBeginCounter[touch_x - 1] = 0;
    touchDown[touch_x - 1] = false;

    // Send Event via SocketIO
    if (sendEvent) {
      sendTouchEvent("touchend", touch_x, getMaxHistoryValue(touch_x, 3), show_JSON_Serial);
    }

    // We expect no other touch events in this run
    return;
  }

  // Touch Move
  // ------------------
  else if ((prevSoftPotValue != softPotValue)) {

    // Handle state machines
    touchDownBegin[touch_x - 1] = false;
    touchDownBeginCounter[touch_x - 1] = 0;

    // check whether the difference exceeds the threshold...
    if (abs(validPreMoveVal - softPotValue) < threshold) return;

    // Serial debug
    if (serialDebug == true) {
      Serial.print("Move from position ");
      Serial.print(validPreMoveVal);
      Serial.print(" to ");
      Serial.println(softPotValue);
    }

    // Temporal store the value for next loop iteration...
    validPreMoveVal = softPotValue;

    // Send Event via SocketIO
    if (sendEvent) {
      sendTouchEvent("touchmove", touch_x, softPotValue, show_JSON_Serial);
    }

    // We expect no other touch events in this run
    return;
  }
}

/*  zeroMinimumValues()
    ------------------
    This function compares three 16 bit intergers
    and zeroes both smaller variables. */
void zeroMinimumValues(int16_t *value1, int16_t *value2, int16_t *value3)
{
  if ((*value3 > *value2) && (*value3 > *value2)) {
    *value2 = 0;
    *value1 = 0;
    return;
  }
  else if ((*value2 > *value3) && (*value2 > *value1)) {
    *value3 = 0;
    *value1 = 0;
    return;
  }
  else if ((*value1 > *value3) && (*value1 > *value2)) {
    *value3 = 0;
    *value2 = 0;
    return;
  }
}

/* cutExtrema()
    ------------------
   This functions cuts a sensor value at the minimum and maximum
*/
void cutExtrema(int16_t *sensorValue, int16_t minimum, int16_t maximum) {
  if (*sensorValue < minimum)
    *sensorValue = minimum;
  if (*sensorValue > maximum)
    *sensorValue = maximum;
}

/* fullRefresh()
    ------------------
*/
void fullRefresh(unsigned char* image) {
  display.setFullWindow();
  display.firstPage();
  do
  {
    display.fillScreen(GxEPD_WHITE);
    display.drawBitmap(0, 0, image, display.epd2.WIDTH, display.epd2.HEIGHT, GxEPD_BLACK);
  }
  while (display.nextPage());
}

/* partialRefresh()
    ------------------
*/
void partialRefresh(unsigned char* image) {
  display.setPartialWindow(0, 0, display.width(), display.height());
  display.firstPage();
  do {
    display.fillScreen(GxEPD_WHITE);
    display.drawBitmap(0, 0, image, display.epd2.WIDTH, display.epd2.HEIGHT, GxEPD_BLACK);


  }
  while (display.nextPage());
}
